

import java.io.*;
import java.util.*;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import topology.*;
import homology.*;
import homology.f2homology.*;

/**
 * This class represents a DOM parser for CityGML data. It computes a finite topological
 * space from the polygons with holes contained in a CityGML document. The vertices of this
 * space are the vertices of the polygons, the lines are derived from the linear rings
 * in the CityGML document, and the faces correspond to the areas of the polygons.
 * Extra faces are given for each interior hole of a polygon given by interior linear rings
 * in the CityGML document. This leads to a finite topological space whose second Betti number
 * equals the number of shells plus the number of interior holes of all individual polygons.
 * By removing these extra faces from the full space, one obtains the "pure" space whose 
 * second Betti number is the number of shells surrounded by polygons with holes
 * in the CityGML document.
 * An Euler characteristic method is inserted in order to allow a fast computation of the second 
 * Betti number after the first and zeroth Betti numbers have been computed.
 * The main method computes the Betti numbers of some example CityGML files contained in the data 
 * folder accompanying this file.
 * 
 * @author Patrick Erik Bradley
 * @version 1.0 --- June 5, 2014
 */
public class CityGMLDomParser {
    
    /**
     * set of faces which correspond to interior holes in polygons
     */
    private static Set<String> extraFaces = new TreeSet();
    
    /**
     * set of faces which correspond to areas of polygons
     */
    private static Set<String> Faces = new TreeSet();

    /**
     * set of lines which connect vertices of polygons
     */
    private static Set<String> Lines = new TreeSet();

    /**
     * set of vertices of polygons
     */
    private static Set<String> Vertices = new TreeSet();

    /**
     * set of 1-dimensional finite topological spaces
     */
    private static Set<FiniteSpace> lineSpaceSet 
	= new TreeSet(new FiniteSpaceComparator());

    /**
     * the CityGML document
     */
    private Document doc;

    /**
     * Parses the CityGML file.
     * @param house the CityGML file
     */
    public CityGMLDomParser(File house) {

	try{
	    DocumentBuilderFactory dbFactory 
		= DocumentBuilderFactory.newInstance();
	    DocumentBuilder dBuilder = dbFactory.newDocumentBuilder();
	    doc = dBuilder.parse(house);
	} catch (Exception ex) {
	    ex.printStackTrace();
	}
    }

    /**
     * Returns the finite topological space associated with a 
     * CityGML document
     */
    public FiniteSpace<String> space() {
	
	doc.getDocumentElement().normalize();
	NodeList polygons = doc.getElementsByTagName("gml:Polygon");
	Set<String> pointSet = new TreeSet();
	BinaryRelation<String> relation = new BinaryRelation();
	for (int i = 0; i < polygons.getLength(); i++) {
	    Node node = polygons.item(i);
	    if (node.getNodeType() == Node.ELEMENT_NODE) {
		FiniteSpace polygon = makePolygon(node);
		pointSet.addAll(polygon.getPointSet());
		relation.addAll(polygon.getRelation().getPairs());
	    }
	}
	FiniteSpace<String> space = new FiniteSpace(pointSet,relation);
	return space;
    }

    /**
     * Returns the finite topological space without the faces
     * associated with the interior holes of polygons. 
     */
    public FiniteSpace<String> pureSpace() {
	FiniteSpace<String> pureSpace = new FiniteSpace(space());
	pureSpace.removeAllPoint(extraFaces);
	return pureSpace;
    }
    
    /**
     * Returns a quick approximation to the core of the finite 
     * topological space associated with this CityGML document
     * by removing beat points.
     */
    public FiniteSpace<String> quasiCore() {
	FiniteSpace<String> core = new FiniteSpace(space());
	for (String line : Lines) {
		if (PointReductions.isUpBeatPoint(core,line)) {
		    core.removePoint(line);
		}
	}
	for (String vertex : Vertices) {
		if (PointReductions.isUpBeatPoint(core,vertex)) {
		    core.removePoint(vertex);
		}
	}
	for (String face : Faces) {
	    if (PointReductions.isDownBeatPoint(core,face)) {
		core.removePoint(face);
	    }
	}
	for (String extraFace : extraFaces) {
	    if (PointReductions.isDownBeatPoint(core,extraFace)) {
		core.removePoint(extraFace);
	    }
	}
    	return core;
    }

    /**
     * Returns a quick approximation to the core of the pure space
     * associated with this CityGML document by removing beat points.
     */
    public FiniteSpace<String> pureQuasiCore() {
	FiniteSpace<String> pureCore = new FiniteSpace(pureSpace());
	for (String line : Lines) {
	    if (PointReductions.isUpBeatPoint(pureCore,line)) {
		pureCore.removePoint(line);
	    }
	}
	for (String vertex : Vertices) {
	    if (PointReductions.isUpBeatPoint(pureCore,vertex)) {
		pureCore.removePoint(vertex);
	    }
	}
	for (String face : Faces) {
	    if (PointReductions.isDownBeatPoint(pureCore,face)) {
		pureCore.removePoint(face);
	    }
	}
	return pureCore;
    }

    /**
     * Returns a polygon as a finite topological space.
     * @param node
     */
    private static FiniteSpace<String> makePolygon(Node node) {
	Set<String> faces = new TreeSet();
	Set<String> eFaces = new TreeSet();
	Set<String> pointSet = new TreeSet();
	BinaryRelation<String> relation = new BinaryRelation();
	String face0 = null;
	NodeList children = node.getChildNodes();
	for (int i = 0; i < children.getLength(); i++) {
	    Node child = children.item(i);
	    String childName = child.getNodeName();
	    if (child.getNodeType() == Node.ELEMENT_NODE) {
		Element element = (Element) child;
		NodeList linearRingList = element.getElementsByTagName("gml:LinearRing");
		for (int j = 0; j < linearRingList.getLength(); j++) {
		    Node linearRing = linearRingList.item(j);
		    String faceId
			= "face_" + linearRing.getAttributes().getNamedItem("gml:id")
			.getNodeValue();
		    if (childName == "gml:exterior") {
			face0 = faceId;
			faces.add(face0);
		    }
		    if (childName == "gml:interior") {
			eFaces.add(faceId);
		    }
		    FiniteSpace circle = makeCircle("gml:pos",linearRing);
		    pointSet.addAll(circle.getPointSet());
		    relation.addAll(circle.getRelation().getPairs());
		    Set<String> circlePoints = circle.KOpenPoints();
		    for (String point : circlePoints) {
			Pair<String,String> pair0 
			    = new ComparablePair(face0,point);
			relation.add(pair0);
			Pair<String,String> pair 
			    = new ComparablePair(faceId,point);
			relation.add(pair);
		    }
		}
	    }
	}
	pointSet.addAll(eFaces);
	pointSet.addAll(faces);
	Faces.addAll(faces);
	extraFaces.addAll(eFaces);
	return new FiniteSpace(pointSet,relation);
    }
    
    /**
     * Returns a circle as a finite topological space.
     * @param tag
     * @param node
     */
    private static FiniteSpace<String> 
	makeCircle(String tag, Node node) {
	
	Set<String> vertices = new TreeSet();
	Set<String> lines = new TreeSet();
	Set<String> pointSet = new TreeSet();
	String nodeId = node.getAttributes().getNamedItem("gml:id")
		.getNodeValue();
	Element element = (Element) node;
	BinaryRelation<String> relation = new BinaryRelation();
	NodeList nodes = element.getElementsByTagName(tag);
	for (int i = 0; i < nodes.getLength()-1; i++) {
	    Node node1 = (Node) nodes.item(i);
	    NodeList nodeList = node1.getChildNodes();
	    Node child = (Node) nodeList.item(0);
	    String childCode = child.getNodeValue();
	    String iString = new Integer(i).toString();
	    String vertex1Id = "vertex(" + childCode + ")";
	    vertices.add(vertex1Id);
	    Node node2 = (Node) nodes.item(i+1);
	    NodeList nodeList2 = node2.getChildNodes();
	    Node child2 = (Node) nodeList2.item(0);
	    String childCode2 = child2.getNodeValue();
	    String vertex2Id = "vertex(" + childCode2 + ")";
	    String lineId = "line" + "_" + nodeId + "_" + iString;
	    for (FiniteSpace lineSpace : lineSpaceSet) {
		if (lineSpace.getPointSet().contains(vertex1Id) 
		    && lineSpace.getPointSet().contains(vertex2Id)) { 
		    Iterator<String> iter = lineSpace.KOpenPoints().iterator();
		    while (iter.hasNext()) {
			lineId = iter.next();
		    }
		    break;
		} else {
		    lineId = "line" + "_" + nodeId + "_" + iString;
		}
	    }
	    
	    lines.add(lineId);
	    Pair<String,String> pair1 
		= new ComparablePair(lineId,vertex1Id);
	    Pair<String,String> pair2
		= new ComparablePair(lineId,vertex2Id);
	    relation.add(pair1);
	    relation.add(pair2);
	    Set<String> lineSet = new TreeSet();
	    lineSet.add(vertex1Id);
	    lineSet.add(vertex2Id);
	    lineSet.add(lineId);
	    BinaryRelation<String> lineRelation = new BinaryRelation();
	    lineRelation.add(pair1);
	    lineRelation.add(pair2);
	    FiniteSpace<String> lineSpace 
		= new FiniteSpace(lineSet,lineRelation);
	    lineSpaceSet.add(lineSpace);
	}
	pointSet.addAll(vertices);
	pointSet.addAll(lines);
	FiniteSpace<String> circle = new FiniteSpace(pointSet,relation);
	Vertices.addAll(vertices);
	Lines.addAll(lines);
	return circle;    
    }
	
    /**
     * Returns the vertices of the space associated
     * with this CityGML document.
     */
    public Set<String> getVertices() {
	return Vertices;
    }

    /**
     * Returns the lines of the space associated with
     * this CityGML document.
     */
    public Set<String> getLines() {
	return Lines;
    }

    /**
     * Returns the faces of the space associated with
     * this CityGML document.
     */
    public Set<String> getFaces() {
	return Faces;
    }

    /**
     * Returns the faces associated to the interior holes
     * of polygons of this CityGML document.
     */
    public Set<String> getExtraFaces() {
	return extraFaces;
    }

    /**
     * Returns the EulerCharacteristic of a chain complex
     * over the finite space F2 with two elements.
     * @param complex the chain complex over F2
     */
    public int EulerCharacteristic(F2ChainComplex complex) {
	List<F2Matrix> boundaryOperator = complex.getBoundaryOperator();
	int dimension = complex.getDimension();
	int EulerCharacteristic = 0;
	for (int i = 0; i < dimension + 1; i++) {
	    int iCells = boundaryOperator.get(i).columns().size();
	    if (i % 2 == 0 % 2) {
		EulerCharacteristic = EulerCharacteristic + iCells;
	    } else {
		EulerCharacteristic = EulerCharacteristic - iCells;
	    }
	}
	return EulerCharacteristic;
    }
    
    
    public static void main(String args[]) {
	
	try {
	    
	    Timer timer = new Timer();
	    
	    // data
	    File house = new File("data/tetrahedron.xml");
	    //File house = new File("data/tetrahedron+hole.xml");
	    //File house = new File("data/meinhaus.xml");
	    //File house = new File("data/meinehaeuser.xml");
	    //File house = new File("data/surface.xml");
	    //File house = new File("data/window.xml");
	    //File house = new File("data/wallSurface.xml");
	    //File house = new File("data/cityObjectMember.xml");
	   
	    // parse
	    timer.start();
	    System.out.println("parsing...");
	    CityGMLDomParser parser = new CityGMLDomParser(house);
	    timer.end();
	    long totalTime1 = timer.getTotalTime();
	    System.out.println("Time: " + timer.timeFormat(totalTime1));
	    System.out.println();

	    // compute space
	    timer.start();
	    System.out.println("computing finite topological space...");
	    FiniteSpace<String> space = parser.space();
	    System.out.print("point space: " + space.getPointSet().size()); 
	    System.out.println(" points = " + Vertices.size() 
	    		       + " vertices + " + Lines.size() + " lines + " 
	    		       + Faces.size() + " faces + "
			       + extraFaces.size() + " extra faces");
	    System.out.println("relation: " + space.getRelation().getPairs().size() + " pairs");
	    timer.end();
	    long totalTime2 = timer.getTotalTime();
	    System.out.println("Time: " + timer.timeFormat(totalTime2));
	    System.out.println();
	   
	    // compute quasi-core
	    timer.start();
	    System.out.println("computing pure quasi-core...");
	    FiniteSpace<String> pureQuasiCore = parser.pureQuasiCore();
	    System.out.println("point space: " + pureQuasiCore.getPointSet().size() + " points");
	    System.out.println("relation: " + pureQuasiCore.getRelation().getPairs().size() + " pairs");
	    //System.out.println(pureCore.getPointSet());
	    timer.end();
	    long totalTime3 = timer.getTotalTime();
	    System.out.println("Time: " + timer.timeFormat(totalTime3));
	    System.out.println();

	    // spanning tree
	    timer.start();
	    System.out.println("Spanning tree...");
	    PrimMST mst = new PrimMST(pureQuasiCore);
	    System.out.print("Pair difference: ");
	    int pairDiff = pureQuasiCore.getRelation().getPairs().size() 
		- mst.getMstRelation().getPairs().size();
	    System.out.println(pairDiff);
	    timer.end();
	    long totalTime4 = timer.getTotalTime();
	    System.out.println("Time: " + timer.timeFormat(totalTime4));
	    System.out.println(); 
	    
	    // quotient complex 
	    timer.start();
	    System.out.println("quotient complex...");
	    FiniteSpace mstSpace 
		= new FiniteSpace(pureQuasiCore.getPointSet(),
				  mst.getMstRelation());
	    F2OrderComplexQuotient pureQuotient = 
		new F2OrderComplexQuotient(mstSpace,pureQuasiCore);
	    timer.end();
	    long totalTime5 = timer.getTotalTime();
	    System.out.println("Time: " + timer.timeFormat(totalTime5));
	    System.out.println();
	    
	    // Betti numbers via quotient of pure space
	    timer.start();
	    System.out.println("Betti numbers...");
	    F2ChainComplex pureQuotientComplex = pureQuotient
		.getF2ChainComplex();
	    int ec = parser.EulerCharacteristic(pureQuotientComplex);
	    int[] pureBetti = pureQuotientComplex.BettiNumbers();
	    System.out.println("b0 = " + (mst.componentNumber()));
	    for (int i = 1; i < pureBetti.length; i++) {
		if (i == 2) {
		    int pureBetti2 = ec + pureBetti[1];
		    System.out.println("b2 = " + (pureBetti2 + extraFaces.size()));
		} else {
		    System.out.println("b" + i + " = " + pureBetti[i]);
		}
	    }
	    for (int i = pureQuasiCore.dimension() + 1; i < space.dimension() + 1; i++) {
		if (i == 2) {
		    System.out.println("b2 = " + extraFaces.size());
		} else {
		    System.out.println("b" + i + " = 0");
		}
	    }
	    timer.end();
	    long totalTime6 = timer.getTotalTime();
	    System.out.println("Time: " + timer.timeFormat(totalTime6));
	    System.out.println();
	    System.out.println("Total time: " + timer.timeFormat(totalTime1 + totalTime2 + totalTime3 + totalTime4 + totalTime5 + totalTime6));
	    
	} catch (Exception ex) {
	    ex.printStackTrace();
	}
    }


}
