import topology.*;
import java.util.*;

/**
 * This class provides for a comparator of finite topological spaces. They are compared with
 * respect to lexicographical orderings of their point sets, and then of their pre-orders. 
 *
 * @author Patrick Erik Bradley
 */
public class FiniteSpaceComparator<E extends Comparable<E>> implements Comparator<FiniteSpace<E>> {

    /**
     * Returns the compare-value of two finite topological spaces with respect to lexicographical
     * ordering: a positive value if the second space is before the first
     * space, a negative value if the space is before the second
     * space, and zero if both are the same.
     * @param <E> space type
     * @param space1 first topological space
     * @param space2 second topological space
     * @return compare-value
     */ 
    @Override
	public int compare(FiniteSpace<E> space1, FiniteSpace<E> space2) {
	TreeSet<E> pointSet1 = new TreeSet();
	TreeSet<E> pointSet2 = new TreeSet();
	TreeSet<Pair<E,E>> relationSet1 = new TreeSet();
	TreeSet<Pair<E,E>> relationSet2 = new TreeSet();
	pointSet1.addAll(space1.getPointSet());
	pointSet2.addAll(space2.getPointSet());
	relationSet1.addAll(space1.getPreOrder().getPairs());
	relationSet2.addAll(space2.getPreOrder().getPairs());
	int result = compare(pointSet1,pointSet2);
	if (result != 0) {
	    return result;
	}
	result = pairCompare(relationSet1,relationSet2);
	return result;
    }
    
    public int compare(TreeSet<E> treeSet1, TreeSet<E> treeSet2) {
	    TreeSet<E> set1 = (TreeSet<E>)treeSet1.clone();
	    TreeSet<E> set2 = (TreeSet<E>)treeSet2.clone();
	    int n = Math.min(set1.size(),set2.size());
	    for (int i = 0; i < n; i++) {
		E e1 = set1.pollFirst();
		E e2 = set2.pollFirst();
		int result = e1.compareTo(e2);
		if (result != 0) {
		    return result;
		}
	    }
	    return set1.size() - set2.size();
    }

    public int pairCompare(TreeSet<Pair<E,E>> treeSet1, TreeSet<Pair<E,E>> treeSet2) {
	TreeSet<Pair<E,E>> set1 = (TreeSet<Pair<E,E>>)treeSet1.clone();
	TreeSet<Pair<E,E>> set2 = (TreeSet<Pair<E,E>>)treeSet2.clone();
	int n = Math.min(set1.size(),set2.size());
	for (int i = 0; i < n; i++) {
	    Pair<E,E> e1 = set1.pollFirst();
	    Pair<E,E> e2 = set2.pollFirst();
	    int result = compare(e1,e2);
	    if (result != 0) {
		return result;
	    }
	}
	return set1.size() - set2.size();
    }

    public int compare(Pair<E,E> pair1, Pair<E,E> pair2) {
	int result = pair1.getKey().compareTo(pair2.getKey());
	if (result != 0) {
	    return result;
	} 
	result = pair1.getValue().compareTo(pair2.getValue());
	return result;
    }
    
}
